// ********************************************************************************** //
// The contents of this file are subject to the Mozilla Public License Version 1.1    //
// (the "License"); you may not use this file except in compliance with the License.  //
// You may obtain a copy of the License at http://www.mozilla.org/MPL/                //
//                                                                                    //
// Software distributed under the License is distributed on an "AS IS" basis,         //
// WITHOUT WARRANTY OF ANY KIND, either express or implied.  See the License for the  //
// specific language governing rights and limitations under the License.              //
//                                                                                    //
// The Original Code provided by Mozilla's Firefox is XUL, Javascript, and CSS.       // 
// The Initial Developer of the Original Code is Aseem Kishore.                       //
// Portions created by Scott Williamson are Copyright (c) 2006.  All Rights Reserved. //
// Contributor - Intuit Inc. (c) 2007                                                 //
// ********************************************************************************** //

// filmstrip.js

var thumbstrips_FS_CHECK_REP = true;

// aSortArray is an array of three boolean elements:
// [ firstVisit?, lastVisit?, everyVisit ]
// exactly one should be true
// send over aNodeList and aListOfIdLists to initialize this filmstrip with
// existing nodes, e.g. for copying a filmstrip, or loading one from a file
// aNodeList is unsorted or arbitrarily sorted, and each element in aNodeList
// should be a reference to an actual WebpageNode object. each node should exist
// exactly once in this list. aListOfIdLists is of the form:
// [ sorted by first visit, sorted by last visit, sorted by every visit ]
// That means the first two lists are of the same length and have exactly one
// instance of each node, and the third list is at least as big. Each element
// in each of the three lists should be the ids for the nodes in aNodeList.
// This trusting architecture was chosen to optimize time performance during
// loading -- don't abuse it! =) Make sure the idLists are sorted correctly.
function thumbstrips_Filmstrip(aSortArray, aNodeList, aListOfIdLists) {
	
	// Representation explanation:
	
	// We store three lists, one for each way to sort the nodes.
	// These lists contain only the ids of the nodes.
	// The actual nodes are stored in a hash map that maps id -> node.
	
	// List 0 is with all the id's sorted by first visit.
	// List 1 is with all the id's sorted by last visit.
	// List 2 is with all the id's sorted by every visit.
	
	// Note that list 0 and list 1 will be the same length as the hash map,
	// but list 2 will be either the same length or greater.
	
	
	// Private fields
	
	var self = this;
	var numNodes = 0;
	var nodeMap = new Array();
	var nodeLists = [ new Array(), new Array(), new Array() ];
	var currentListIndex = 1;
	
	// if we have aSortArray, parse it
	
	if (aSortArray) {
		thumbstrips_checkArrayType("filmstrip", aSortArray);
		thumbstrips_checkTrue("filmstrip", aSortArray.length === 3, "aSortArray.length != 3");
		
		for (var i = 0; i < 3; i++) {
			thumbstrips_checkType("filmstrip", aSortArray[i], "boolean");
			
			if (aSortArray[i] === true) {
				currentListIndex = i;
			}
		}
	}
	
	// if we have aNodeList and aListOfIdLists, initialize everything
	
	if (aNodeList && aListOfIdLists) {
		thumbstrips_checkArrayType("filmstrip", aNodeList);
		
		// initialize numNodes and nodeMap
		
		for (var i = 0; i < aNodeList.length; i++) {
			var node = aNodeList[i];
			
			thumbstrips_checkWebpageNodeType("filmstrip", node);
			
			var index = hash(node);
			
			nodeMap[index] = node; // we don't care about overwriting
			node.id = index;
		}
		
		numNodes = thumbstrips_propLength(nodeMap);
		
		thumbstrips_checkArrayType("filmstrip", aListOfIdLists);
		thumbstrips_checkTrue("filmstrip", aListOfIdLists.length === 3, "aListOfIdLists.length != 3");
		thumbstrips_checkArrayType("filmstrip", aListOfIdLists[0]);
		thumbstrips_checkArrayType("filmstrip", aListOfIdLists[1]);
		thumbstrips_checkArrayType("filmstrip", aListOfIdLists[2]);
		thumbstrips_checkTrue("filmstrip", aListOfIdLists[0].length === numNodes, "aListOfIdLists[0].length != " + numNodes);
		thumbstrips_checkTrue("filmstrip", aListOfIdLists[1].length === numNodes, "aListOfIdLists[1].length != " + numNodes);
		thumbstrips_checkTrue("filmstrip", aListOfIdLists[2].length >= numNodes, "aListOfIdLists[2].length < " + numNodes);
		
		nodeLists[0] = thumbstrips_copy(aListOfIdLists[0]);
		nodeLists[1] = thumbstrips_copy(aListOfIdLists[1]);
		nodeLists[2] = thumbstrips_copy(aListOfIdLists[2]);
	}
	
	
	
//	// if we have aNodeList, then initialize nodeMap
//	
//	if (aNodeList) {
//		for (var i =0; i < nodeList.length; i++) {
//			var node = nodeList[i];
//			var index = hash(node);
//			
//			if (nodeMap[index]) { // a mapping already exists with this hash
//				// uh oh! clear everything, just make a new filmstrip
//				nodeMap = new Array();
//				nodeList = new Array();
//				break;
//			}
//			
//			nodeMap[index] = node;
//			node.id = index;
//		}
//	}
	
	
	// Public popeties
	
	// [none]
	
	
	// Private helpers
	
	function checkRep() {
		if (thumbstrips_FS_CHECK_REP === false) {
			return;
		}
		
		var cr = self.toString() + " checkRep";
		var i;
		
		// numNodes is a number
		thumbstrips_checkType(cr, numNodes, "number");
		
		// currentListIndex is a number
		thumbstrips_checkType(cr, currentListIndex, "number");
		
		// 0 <= currentListIndex < nodeLists.length
		thumbstrips_checkTrue(cr, 0 <= currentListIndex, "currentListIndex < 0");
		thumbstrips_checkTrue(cr, currentListIndex < nodeLists.length, "currentListIndex >= nodeLists.length");
		
		// nodeMap is an array
		thumbstrips_checkArrayType(cr, nodeMap);
		
		// nodeLists is an array
		thumbstrips_checkArrayType(cr, nodeLists);
		
		// each nodeList is an array
		thumbstrips_checkArrayType(cr, nodeLists[0]);
		thumbstrips_checkArrayType(cr, nodeLists[1]);
		thumbstrips_checkArrayType(cr, nodeLists[2]);
		
		// nodeMap.length = 0
		thumbstrips_checkEquals(cr, nodeMap.length, 0, "nodeMap.length != 0");
		
		// for each node in nodeMap, node is a webpageNode, and index of node = node.id
		for (i in nodeMap) {
			thumbstrips_checkWebpageNodeType(cr, nodeMap[i]);
			thumbstrips_checkEquals(cr, i, nodeMap[i].id, "nodeMap[" + i + "].id unequal, id = " + nodeMap[i].id);
		}
		
		// propLength of nodeMap = nodeLists[0].length and nodeLists[1].length
		thumbstrips_checkEquals(cr, thumbstrips_propLength(nodeMap), nodeLists[0].length, "propLength of nodeMap != nodeLists[0].length");
		thumbstrips_checkEquals(cr, thumbstrips_propLength(nodeMap), nodeLists[1].length, "propLength of nodeMap != nodeLists[1].length");
		
		// nodeLists[2].length >= nodeLists[0].length
		thumbstrips_checkTrue(cr, nodeLists[2].length >= nodeLists[0].length, "nodeLists[2].length < nodeLists[0].length");
	}
	
	function hash(node) {
		return node.url;
	}
	
	function nodeComparison(node1) {
//		return function(node2) {
//			return node1.id === node2.id;
//		}
//		return function(node2) {
//			return node1 === node2;
//		}
		return function(idOfNode2) {
			return node1.id === idOfNode2;
		}
	}
	
	function indicesToNodes(listIndex, debug) {
//		if (!listIndex) {
//			listIndex = currentListIndex;
//		}
		
		var array = new Array();
		var ids = nodeLists[listIndex];
		
		if (debug) {
			thumbstrips_log("lI = " + listIndex);
			thumbstrips_log("ids = " + ids);
		}
		
		for (var i = 0; i < ids.length; i++) {
			array[i] = nodeMap[ids[i]];
		}
		
		if (debug) {
			thumbstrips_log("array = " + array);
		}
		
		return array;
	}
	
	
	// Public methods
	
	this.getNodes = function() {
		// need to make a new array because nodeLists arrays only contain ids
		// and we need to return actual nodes
		
		return indicesToNodes(currentListIndex);
	};
	
	this.getNodeById = function(id) {
		return nodeMap[id];
	};
	
	this.getNumUniqueNodes = function() {
		return numNodes;
	};
	
	this.getCurrentSorting = function() {
		switch (currentListIndex) {
			case 0:
				return "first visit";
			case 1:
				return "last visit";
			case 2:
				return "every visit";
			default:
				return "ERROR, cLI = " + currentListIndex;
		}
	};
	
	this.getNodesByDifferentSorting = function(tempSortArray) {
		thumbstrips_checkArrayType("getDifferentSorting", tempSortArray);
		thumbstrips_checkTrue("getDifferentSorting", tempSortArray.length === 3, "aSortArray.length != 3");
		
		for (var i = 0; i < 3; i++) {
			thumbstrips_checkType("getDifferentSorting", tempSortArray[i], "boolean");
			
			if (tempSortArray[i] === true) {
//				thumbstrips_log("sending i = " + i);
				return indicesToNodes(i);
			}
		}
	};
	
	this.addOrUpdateNode = function(title, url, domain, init) {
		var node = new thumbstrips_WebpageNode(title, url, domain);
		var index = hash(node);
		
		if (nodeMap[index]) {	// node already exists
		
			// update it
			node = nodeMap[index];
//			node.title = title;
//			node.url = url;
//			node.domain = domain;
			
			// we don't have to do anything to the 0 (first visit) list
			// because it's original first visit hasn't changed
			
			// for the 1 (last visit) list, remove the node from the previous
			// position and add it to the end of the list
			
			var i = thumbstrips_reverseLinearSearch(nodeLists[1], nodeComparison(node));
			if (i === -1) {
				thumbstrips_log("Error, nodeMap[" + index + "] = " + node + ", but not found in nodeLists[1]!");
			} else {
				nodeLists[1].splice(i,1);
			}
		
			nodeLists[1][nodeLists[1].length] = index;
			
			// for the 2 (every visit) list, first check if there is already a
			// copy at the end -- in that case, don't add a new copy -- but if
			// not, add a copy at the end. also, begin a new visit for that node
			// if a copy doesn't exist at the end, otherwise update the current.
			
			if (nodeLists[2][nodeLists[2].length - 1] === index) {
				
				// a copy already exists at the end
				
				node.updateCurrentVisit(true);
				
			} else {
				
//				thumbstrips_log("no copy exists, last id = " + nodeLists[2][nodeLists[2].length - 1].id + ", index = " + index);
				
				// no copy exists, add one
				
				node.beginNewVisit();
				nodeLists[2][nodeLists[2].length] = index;
				
			}
			
//			node.beginNewVisit();
			
//			if (sortArray[0] === true) { // sorting by FIRST VISIT
//				
//				// don't move it in the list at all
//					
//			} else if (sortArray[1] === true) { // sorting by LAST VISIT
//			
//				// remove node from previous position...
//				var i = thumbstrips_reverseLinearSearch(nodeList, nodeComparison(node));
//				if (i === -1) {
//					alert("Error, nodeMap[" + index + "] = " + node + ", but not found in nodeList!");
//				} else {
//					nodeList.splice(i,1);
//				}
//			
//				// ...and add it to the end
//				nodeList[nodeList.length] = node;
//				
//			} else { // sorting by EVERY VISIT
//			
//				// just add a copy to the end
//				nodeList[nodeList.length] = node;
//				
//			}
			
		} else {	// node doesn't exist
		
			// increase count
			numNodes++;
		
			// initialize it
			nodeMap[index] = node;
			node.id = index;
			
			if (init && (typeof init) === "function") {
				init(node);
			}
			
			// add it to the end of every list
			
//			thumbstrips_log("index = " + index);
			
			nodeLists[0][nodeLists[0].length] = index;
			nodeLists[1][nodeLists[1].length] = index;
			nodeLists[2][nodeLists[2].length] = index;
			
//			// add it to the end
//			nodeList[nodeList.length] = node;
		
		}
		
		checkRep();
		
//		thumbstrips_log("l0 = " + nodeLists[0]);
//		thumbstrips_log("l1 = " + nodeLists[1]);
//		thumbstrips_log("l2 = " + nodeLists[2]);
		
		return node;
	};
	
	// if the filmstrip is being sorted by EVERY VISIT, then the user should
	// have a choice between deleting a single count of a website versus the
	// entire node encapsulating that website. so in this case, send count
	// (some value between 0 and node.visitCount - 1) to delete a single count,
	// and leave it out to delete the entire node.
	this.removeNode = function(node, count) {
		thumbstrips_checkWebpageNodeType("removeNode", node);
		
		if (count) {
			thumbstrips_checkType("removeNode", count, "number");
			thumbstrips_checkTrue("removeNode", 0 <= count, "count < 0");
			thumbstrips_checkTrue("removeNode", count < node.getVisitCount(), "count >= node.visitCount");
		}
		
		var other = nodeMap[node.id];
		
		if (!other) {
			// node is not in the filmstrip
			return;
		}
		
		// if no count given, or if the node only has 1 visit
		if (!count || node.getVisitCount() === 1) {
		
			// delete the actual node
			delete nodeMap[node.id];
			
			var index0 = thumbstrips_linearSearch(nodeLists[0], nodeComparison(node));
			var index1 = thumbstrips_linearSearch(nodeLists[1], nodeComparison(node));
			var index2 = thumbstrips_linearSearch(nodeLists[2], nodeComparison(node));
			
			// remove the single id from both 0 and 1 lists
			nodeLists[0].splice(index0, 1);
			nodeLists[1].splice(index1, 1);
			
			// remove every instance of the id from the 2 list
			while (index2 >= 0) {
				nodeLists[2].splice(index2, 1);
				index2 = thumbstrips_linearSearch(nodeLists[2], nodeComparison(node));
			}
			
		} else { // we are only deleting a single instance
			
			// don't touch the 0 and 1 lists or the actual node
			// instead, only delete the appropriate id from the 2 list
			
			var index = -1;
			var counter = -1;
			
			for (var index = 0; index < nodeLists[2].length; index++) {
				
				if (nodeComparison(node)(nodeLists[2][index]) === true) {
					counter++;
					
					if (count === counter) {
						break;
					}
				}
				
			}
			
			nodeLists[2].splice(index, 1);			
		
		}
		
//		if (sortArray[2] === false) { // if we're NOT sorting by every visit
//		
//			delete nodeMap[node.id];
//		
//			var index = thumbstrips_linearSearch(nodeList, nodeComparison(node));
//		
//			nodeList.splice(index, 1);
//			
//		} else { // if we ARE sorting by every visit
//		
//			if (!count) { // delete the entire node
//				
//				delete nodeMap[node.id];
//				
//				// go through list and delete every instance of node
//				
//				var index = thumbstrips_linearSearch(nodeList, nodeComparison(node));
//				
//				while (index >= 0) {
//					nodeList.splice(index, 1);
//					index = thumbstrips_linearSearch(nodeList, nodeComparison(node))
//				}
//				
//			} else { // delete only a single instance
//			
//				var index = -1;
//				var counter = -1;
//				
//				for (var index = 0; index < nodeList.length; index++) {
//					
//					if (nodeComparison(node)(nodeList[index]) === true) {
//						counter++;
//						
//						if (count === counter) {
//							break;
//						}
//					}
//					
//				}
//				
//				nodeList.splice(index, 1);
//				
//			}
//			
//		}
		
		checkRep();
	};
	
	this.sortByFirstVisit = function() {
		currentListIndex = 0;
	};
	
	this.sortByLastVisit = function() {
		currentListIndex = 1;
	};
	
	this.sortByEveryVisit = function() {
		currentListIndex = 2;
	};
	
	this.toString = function() {
		return "Filmstrip [" + self.getNodes() + "]";
	};
	
	checkRep();
}